#!/bin/sh
#
# @(#)phatsort.sh	1.8 06/01/08
#
# Linux/Unix shell script to fix the PhatNoise(R) Phatbox
# DMS database and index files for improved sorting.
#
# by Ti Kan (http://www.amb.org/ti)
# based on a similar script by "Cowboy" Ben Alman (cowboy@rj3.net)
#
# See the accompanied README file for information.
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# This is needed on some platforms or else sort(1) would infinite loop
LANG=C; export LANG

# Pid of this process
pid=$$

# Command line argument check
if [ $# != 1 ]
then
	echo "Usage: $0 <phtdata>" >&2
	echo "where <phtdata> is the mount point of your DMS phtdta partition" >&2
	exit 1
fi

phtdta="$1"

# Sanity check
if [ ! -d $phtdta ]
then
	echo "$phtdta is not a directory!" >&2
	exit 2
fi
if [ ! -w $phtdta ]
then
	echo "You do not have write permission at $phtdta" >&2
	exit 2
fi
if [ ! -f ${phtdta}/artist.db -o ! -f ${phtdta}/album.db -o \
     ! -f ${phtdta}/genre.db  -o ! -f ${phtdta}/tracks.db ]
then
	echo "${phtdta}: Missing required database files." >&2
	exit 3
fi

# Check for required external programs
found_build_idx=0
found_unix2dos=0
dirs=`echo $PATH | sed 's/:/ /g'`
for p in $dirs
do
	if [ -x ${p}/build_idx ]
	then
		found_build_idx=1
	fi
	if [ -x ${p}/unix2dos ]
	then
		found_unix2dos=1
	fi
done

if [ $found_build_idx -eq 0 ]
then
	echo "Missing required program \"build_idx\"." >&2
	echo "Please compile build_idx.c program and install it" >&2
	echo "somewhere in your PATH." >&2
	exit 4
fi
if [ $found_unix2dos -eq 0 ]
then
	echo "Missing required program \"unix2dos\"." >&2
	echo "Please install the appropriate OS package for this utility." >&2
	exit 4
fi

echo "Backing up original db and index files (phatsort_bak_${pid}.tgz) ..."
files="artist.db album.db genre.db tracks.db"
for f in tracks.idx tracks.dat
do
	if [ -f $f ]
	then
		$files="$files $f"
	fi
done
for d in Profiles
do
	if [ -d $d ]
	then
		$files="$files $d"
	fi
done
(cd ${phtdta}; tar czf - $files) >phatsort_bak_${pid}.tgz
if [ $? -ne 0 ]
then
	echo "Aborted: backup failed." >&2
	exit 5
fi

trap "rm -f rename_tts.sh /tmp/phatsort_tmp?.${pid}" 1 2 3 5 15

# Read original db file
echo "Reading the original db files ..."
cat ${phtdta}/tracks.db | \
sed "s/^\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\(.*\)$/\3\t\3\t\1\t\2\t\4\t\5\t\6\t\7\t\8/" | \
sed "s/^The The\b/__tHe_tHe__/; s/^The \(.*\)\t/\1\t/; s/^A \(.*\)\t/\1\t/; s/^An \(.*\)\t/\1\t/; s/__tHe_tHe__/The The/" \
	>/tmp/phatsort_tmp1.${pid}
if [ $? -ne 0 ]
then
	echo "Aborted: db file parse failure." >&2
	exit 6
fi

cat /tmp/phatsort_tmp1.${pid} | \
sed "s/^\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\([^\t]*\)\t\(.*\)$/\3\t\4\t\1\t\5\t\6\t\7\t\8\t\9/" \
	>/tmp/phatsort_tmp2.${pid}
if [ $? -ne 0 ]
then
	echo "Aborted: db file parse failure." >&2
	exit 7
fi

# Create tmp voice file rename script
cat /tmp/phatsort_tmp1.${pid} | \
sed "s	^\([^\t]*\)\t\([^\t]*\)\t.*$	cp \"${phtdta}\/tts\/\2.wav\" \"${phtdta}\/tts\/\1.wav\"	" \
>/tmp/phatsort_tmp3.${pid}
sort -f -b -d /tmp/phatsort_tmp3.${pid} | uniq >rename_tts.sh

# Sort the db files
echo "Re-sorting the db files ..."
sort -f -b -d -t '	' -k 2,2 \
	/tmp/phatsort_tmp2.${pid} >${phtdta}/tracks.db
sort -f -b -d -t '	' -k 3,3 -k 4,4 -k 8,8n -k 7,7 \
	/tmp/phatsort_tmp2.${pid} >${phtdta}/artist.db
sort -f -b -d -t '	' -k 4,4 -k 8,8n -k 3,3 -k 7,7 \
	/tmp/phatsort_tmp2.${pid} >${phtdta}/album.db
sort -f -b -d -t '	' -k 5,5 -k 3,3 -k 4,4 -k 8,8n -k 7,7 \
	/tmp/phatsort_tmp2.${pid} >${phtdta}/genre.db

# Make sure the db files are in DOS format
(cd ${phtdta}; unix2dos tracks.db artist.db album.db genre.db)

echo "Fixing voice prompt files ..."
if [ ! -d ${phtdta}/tts ]
then
	mkdir ${phtdta}/tts
fi
sh ./rename_tts.sh 2>/dev/null && rm -f rename_tts.sh

echo "Removing old index files ..."
if [ ! -d ${phtdta}/Profiles ]
then
	mkdir ${phtdta}/Profiles
fi
if [ ! -d ${phtdta}/Profiles/Default ]
then
	mkdir ${phtdta}/Profiles/Default
fi
(cd ${phtdta}; find Profiles -type f -name '*.idx' -print | xargs rm -f)

echo "Generating new index files (log output: phatsort_idx_${pid}.log)..."
build_idx ${phtdta} >phatsort_idx_${pid}.log 2>&1

rm -f /tmp/phatsort_tmp?.${pid} rename_tts.sh

if [ -r $HOME/.phatsort_ext ]
then
	echo "Running user post-processing script..."
	. $HOME/.phatsort_ext
fi

echo "Done."
exit 0

